/*
 * $Log: handler.c,v $
 * Revision 1.6  1995/08/22  22:30:04  cthiru
 * Changed process*, generator fns and _kqml_handler to use KQML_KMSG_LIST
 *
 * Revision 1.5  1995/08/10  18:42:46  robin
 * New merged Loral/UMBC code.
 *
 * Revision 1.5 1995/06/08   09:58:29  chelliah
 * Changed ftime calls to gettimeofday for SYSV compatibility
 *
 * Revision 1.4  1993/05/10  17:58:47  robin
 * Added processing to add facilitator timings to overrall timings.
 *
 * Revision 1.3  1993/04/30  19:27:39  fritzson
 * Corrected to look for negative values returned from parse_kqml_stream.
 *
 * Revision 1.2  1993/04/25  07:51:16  robin
 * Stable with SAS KRIL.
 *
 * Revision 1.1  1993/01/25  13:47:25  fritzson
 * Initial revision
 *
 *
 */

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>

#include <sys/types.h>
#include <sys/time.h>

#include <regex.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

/*
 *	(c) Copyright 1993 Loral Defense Systems-Eagan
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */

struct monitor_request {
	char *sender;		/* the agent that initiated the monitor request, and to
								which all the responses should be sent */
	char *reply_with;	/* the value of the :reply-with field from the 
								monitoring agent's request */
	char *exp;			/* regular expression sent by the monitoring agent that
								must match information produced by this agent for
								that informationto qualify as information requested */
	struct re_pattern_buffer bufp;	/* compiled form of the above exp */
	struct monitor_request *match_node, *next_node, *prev_node;
	};

struct generator {
	char *client;				/* agent that initiated creation of the generator */
	char *reply_with;			/* the :reply-with field from the client's msg */
	char *query;				/* query that may be used to generate the next
										reply, or replies, from the appln. value of
										this field will be NULL when the appln. has
										provided all the replies. */
	KQML_KMSG_LIST *result; /* list of kqml_messages - doubly linked list */
	long created;				/* Ofcourse, Created time */
	long last_update;			/* last accessed time */
	struct generator *next_node, *prev_node;
	};

static struct monitor_request *mon_req = NULL;
static struct generator *generator = NULL;

static struct monitor_request *find_open_mon_slot (void);
static void free_monitor_slot (struct monitor_request *node);

static struct generator *find_gen_slot (char *key);
static struct generator *find_open_gen_slot (void);
static void free_generator_slot (struct generator *node);

static void process_monitor (kqml_message *kmsg);
static void process_generator (handler hfunction, kqml_message *kmsg);
static void process_next (kqml_message *kmsg);
static void process_rest (kqml_message *kmsg);
static void process_discard (kqml_message *kmsg);
static void process_and_reply (handler hfunction, kqml_message *kmsg);

void kqml_prep_outmsg (kqml_message *msg, kqml_message *outmsg);
static void error_reply (kqml_message *msg, char *message);


int
_kqml_get_subscriber_list (char *content, int content_len,
	char ***subscriber, char ***subscriber_id)
{
	struct monitor_request *match_node, *cur_slot;
	int slot_count;

	*subscriber = *subscriber_id = NULL;
	if (mon_req == NULL)
	{
		return 0;
	}

	slot_count = 0;
	match_node = NULL;
	cur_slot = mon_req->prev_node;
	do
	{
		if (re_match (&cur_slot->bufp, content, content_len, 0, NULL) >= 0)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Monitor: Matching %s vs %s",
				content, cur_slot->exp);
			slot_count++;
			cur_slot->match_node = match_node;
			match_node = cur_slot;
		}
		cur_slot = cur_slot->prev_node;
	}
	while (cur_slot != mon_req->prev_node);

	if (slot_count == 0)
	{
		return 0;
	}

	*subscriber = malloc (sizeof (char *) * slot_count);
	*subscriber_id = malloc (sizeof (char *) * slot_count);

	cur_slot = match_node;
	slot_count = 0;
	while (cur_slot)
	{
		*subscriber[slot_count] = cur_slot->sender;
		*subscriber_id[slot_count] = cur_slot->reply_with;
		slot_count++;
		cur_slot = cur_slot->match_node;
	}
		
	return slot_count;
}

/*
 * This is a utility used for monitor processing.  It will find an 
 * available slot in the table of monitor requests for the addition of a 
 * new request.
*/
static struct monitor_request *
find_open_mon_slot ()
{
	struct monitor_request *new_node;

	new_node = malloc (sizeof (struct monitor_request));
	if (new_node == NULL)
	{
		return NULL;
	}
	new_node->sender = 
		new_node->reply_with =
		new_node->exp = NULL;
	new_node->match_node = NULL;

	if (mon_req == NULL)
	{
		new_node->next_node = 
			new_node->prev_node = 
			mon_req = new_node;
	}
	else
	{
		new_node->next_node = mon_req;
		new_node->prev_node = mon_req->prev_node;
		mon_req->prev_node->next_node = new_node;
		mon_req->prev_node = new_node;
	}

	return new_node;
}

static void
free_monitor_slot (struct monitor_request *node)
{
	if (node)
	{
		if (node->sender)
		{
			free (node->sender);
		}
		if (node->reply_with)
		{
			free (node->reply_with);
		}
		if (node->exp)
		{
			free (node->exp);
		}
		if (node->next_node == node)
		{
			/* This is the only node */
			mon_req = NULL;
		}
		else if (node == mon_req)
		{
			/* If this is the root node, move root */
			mon_req = mon_req->next_node;
		}
		node->prev_node->next_node = node->next_node;
		node->next_node->prev_node = node->prev_node;
		free (node);
	}
}

static struct generator *
find_open_gen_slot ()
{
	struct generator *new_node;

	new_node = malloc (sizeof (struct generator));
	if (new_node == NULL)
	{
		return NULL;
	}

	new_node->client =
		new_node->reply_with =
		new_node->query = NULL;

	new_node->result = NULL;

	new_node->created = 
		new_node->last_update = time (NULL);

	if (generator == NULL)
	{
		new_node->next_node=
			new_node->prev_node = 
			generator = new_node;
	}
	else
	{
		new_node->next_node = generator;
		new_node->prev_node = generator->prev_node;
		generator->prev_node->next_node = new_node;
		generator->prev_node = new_node;
	}
	
	return new_node;
}

static void
free_generator_slot (struct generator *slot)
{
	if (slot)
	{
		if (slot->reply_with)
		{
			free (slot->reply_with);
		}
		if (slot->client)
		{
			free (slot->client);
		}
		if (slot->query)
		{
			free (slot->query);
		}
		if (slot->result)
		{
			kqml_free_kmsg_list (slot->result);
		}

		if (slot == slot->next_node)
		{
			/* This is the only node */
			generator = NULL;
		}
		else if (slot == generator)
		{
			/* This is the root. Move the root forward */
			generator = generator->next_node;
		}
		slot->next_node->prev_node = slot->prev_node;
		slot->prev_node->next_node = slot->next_node;
		free (slot);
	}
}

static struct generator *
find_gen_slot (char *key)
{
  	struct generator *node;

	if (generator)
	{
		node = generator;
		do
		{
			if (kqml_stricmp (node->reply_with, key) == 0)
			{
				return node;
			}
			else
			{
				node = node->next_node;
			}
		}
		while (node != generator);
	}

	return NULL;
}

/* 
 * This is the default handler for monitor performatives.
 * It puts the content field in the table for regexp matching.
 */
static void
process_monitor (kqml_message *kmsg)
{
	kqml_message *outmsg;
	char *reply_with, *sender, *content;
	char *error_msg;
	struct monitor_request *slot;

	reply_with = _kqml_get_reply_with_field (kmsg);
	sender = kqml_get_field (":sender", kmsg, NULL, NULL);
	content = kqml_get_field (":content", kmsg, NULL, NULL);

	if (reply_with == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Monitor cannot be processed without reply-with");
		return ;
	}
	if (sender == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Monitor cannot be processed without sender");
		error_reply (kmsg, "sender not defined");
		return ;
	}
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Monitor cannot be processed without content");
		error_reply (kmsg, "content not defined");
		return ;
	}

	slot = find_open_mon_slot ();
	if (slot == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("No slots available - ignoring monitor request");
		error_reply (kmsg, "not enough resource");
		return ;
	}
  
	/* Enter information into the array holding monitor request information */
	slot->sender = strdup (sender);
	slot->reply_with = strdup (reply_with);
	slot->exp = strdup (content);

	if (slot->sender == NULL || slot->reply_with == NULL ||
		slot->exp == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Not enough memory - ignoring monitor request");
		free_monitor_slot (slot);
		error_reply (kmsg, "not enough memory for monitor info");
		return ;
	}

	/* compile the regular expression passed in */
	error_msg = re_compile_pattern (slot->exp, strlen(slot->exp), &slot->bufp);
	if (error_msg != NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Unable to compile pattern for monitor request: %s", error_msg);
		free_monitor_slot (slot);
		error_reply (kmsg, error_msg);
		return ;
	}

	outmsg = kqml_build_msg ("reply", "OK", slot->sender, NULL, NULL);
	if (outmsg == NULL)
	{
		free_monitor_slot (slot);
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Could not compose reply to monitor.");
		error_reply (kmsg, "not enough memory");
	}
	else
	{
		kqml_prep_outmsg (kmsg, outmsg);

		/* send out the reply */
		_kqml_send_xdr_msg (outmsg);
      
		/* free the outgoing message */
		kqml_free_msg (outmsg);
	}
  
	return ;
}

static void
process_generator (handler hfunction, kqml_message *kmsg)
{
	KQML_REPLY_TYPE reply_type;
	void *reply;
	kqml_message *outmsg;
	char *reply_with, *sender, *content, endOfGenerator;
	struct generator *slot;

	reply_with = _kqml_get_reply_with_field (kmsg);
	sender = kqml_get_field (":sender", kmsg, NULL, NULL);
	content = kqml_get_field (":content", kmsg, NULL, NULL);

	if (reply_with == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Generator cannot be processed without reply-with");
		return ;
	}
	if (sender == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Generator cannot be processed without sender");
		error_reply (kmsg, "sender not defined");
		return ;
	}
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Generator cannot be processed without content");
		error_reply (kmsg, "content not defined");
		return ;
	}

	slot = find_open_gen_slot ();
	if (slot == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Unable to find an open generator slot");
		error_reply (kmsg, "not enough resource");
		return ;
	}

	/* Now build the generator */
	slot->client = strdup (sender);
	slot->reply_with = strdup (reply_with);
	slot->query = strdup (content);

	if (slot->client == NULL || slot->reply_with == NULL
		|| slot->query == NULL )
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for generator slot");
		free_generator_slot (slot);
		error_reply (kmsg, "not enough memory for generator info");
		return ;
	}

	/* 
	 *	Save the connection field. 
	 * Invoke user handle. 
	 * Restore connection field
	*/
	{
		int save_connection;

		save_connection = kmsg->connection;
		reply_type = KQML_REPLY_ERROR;
		reply = hfunction (slot->query, kmsg, &reply_type);
		slot->result = kqml_wrapper (reply_type, reply);
		kmsg->connection = save_connection;
	}

	/* Now format and send out the reply to the NEXT */
	if (slot->result == KQML_HANDLER_NONE || slot->result == KQML_HANDLER_ERROR)
	{
		outmsg = kqml_build_msg ("sorry", "Unable to make generator", 
			slot->client, reply_with, NULL);
		slot->result = KQML_HANDLER_NONE;
		endOfGenerator = TRUE;
	}
	else
	{
		outmsg = kqml_build_msg ("ready", "OK", slot->client, reply_with, NULL);
		endOfGenerator = FALSE;
	}

	if (outmsg)
	{
		kqml_prep_outmsg (kmsg, outmsg);
		_kqml_send_xdr_msg (outmsg);
      
		/* free the outgoing message */
		kqml_free_msg (outmsg);
	}
	else
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Could not compose reply to generator.");
		error_reply (kmsg, "not enough memory");
	}

	if (endOfGenerator)
	{
		free_generator_slot (slot);
	}

	return ;
}

static void
process_next (kqml_message *kmsg)
{
	KQML_KMSG_LIST *temp;
	kqml_message *outmsg;
	struct generator *slot;
	char *gen_key, endOfGenerator;

	/* process directly from the generator */
	gen_key = kqml_get_field (":in-reply-to", kmsg, NULL, NULL);
	if (gen_key == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Next received without in-reply-to");
		return ;
	}

	slot = find_gen_slot(gen_key);
	if (slot == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Unable to find generator requested for key '%s'", gen_key);
		return ;
	}

	/* Now format and send out the reply to the NEXT */
	if (slot->result != KQML_HANDLER_NONE)
	{
		outmsg = slot->result->kmsg;
		temp = slot->result;
		if (slot->result == slot->result->next)
		{
			/* If this is the last message 
				If this a sorry message, we can finish now
				else we have to wait for the next request, send
				a sorry and finish
			*/
			if ( kqml_stricmp (kqml_get_performative (outmsg), "sorry") == 0)
			{
				endOfGenerator = TRUE;
			}
			slot->result = KQML_HANDLER_NONE;
		}
		else
		{
			slot->result->prev->next = slot->result->next;
			slot->result->next->prev = slot->result->prev;
			slot->result = slot->result->next;
		}
		free (temp);
		endOfGenerator = FALSE;
	}
	else
	{
		outmsg = kqml_build_msg ("sorry", NULL, NULL, NULL, NULL);
		endOfGenerator = TRUE;
	}

	/* We should try to use the orignial generator's :reply-with */
	/* If this message has a :reply-with, kqml_prep_outmsg will use it */
	kqml_put_field (":in-reply-to", outmsg, slot->reply_with, -1);
	kqml_prep_outmsg (kmsg, outmsg);

	/* send out the reply */
	_kqml_send_xdr_msg (outmsg);

	/* free the outgoing message */
	kqml_free_msg (outmsg);

	if (endOfGenerator)
	{
		free_generator_slot (slot);
	}
  
	return ;
}

static void
process_rest (kqml_message *kmsg)
{
  KQML_KMSG_LIST *temp;
  kqml_message *outmsg;
  struct generator *slot;
  char *gen_key;
  
	/* process directly from the generator, as a 'stream-all' */
	gen_key = kqml_get_field(":in-reply-to", kmsg, NULL, NULL);
	if (gen_key == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Rest received without in-reply-to");
		return ;
	}

	slot = find_gen_slot(gen_key);
	if (slot == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Unable to find generator requested for key '%s'", gen_key);
		return ;
	}

	while (slot->result != KQML_HANDLER_NONE)
	{
		outmsg = slot->result->kmsg;
		temp = slot->result;
		if (slot->result == slot->result->next)
		{
			slot->result = KQML_HANDLER_NONE;
		}
		else
		{
			slot->result->prev->next = slot->result->next;
			slot->result->next->prev = slot->result->prev;
			slot->result = slot->result->next;
		}
		free (temp);
		kqml_put_field (":in-reply-to", outmsg, slot->reply_with, -1);
		kqml_prep_outmsg (kmsg, outmsg);
		_kqml_send_xdr_msg (outmsg);
		kqml_free_msg (outmsg);
	}

	/* and then send a 'SORRY' to let the client know there's no more */
	outmsg = kqml_build_msg ("sorry", NULL, NULL, NULL,
		":in-reply-to", slot->reply_with, NULL);
	if (outmsg)
	{
		kqml_prep_outmsg (kmsg, outmsg);
		_kqml_send_xdr_msg (outmsg);
		kqml_free_msg (outmsg);
	}

	free_generator_slot (slot);

	return ;
}

static void
process_discard (kqml_message *kmsg)
{
	struct generator *slot;
	char *gen_key;
  
  /* process directly from the generator */
	gen_key = kqml_get_field(":in-reply-to", kmsg, NULL, NULL);
	if (gen_key == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Discard received without in-reply-to");
		return ;
	}

	slot = find_gen_slot(gen_key);
	if (slot == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("Unable to find generator requested for key '%s'", gen_key);
		return ;
	}

	free_generator_slot (slot);

	return ;
}


/*
 *  process_and_reply
 *
 *  Default processing for a performative, which is to find a handler function
 *  for the performative, invoke that function and send a KQML 'reply' back to
 *  the client that contains, as it's content, the result passed back from the 
 *  handler function.
 *
 */
static void
process_and_reply (handler hfunction, kqml_message *kmsg)
{
	KQML_REPLY_TYPE reply_type;
	void *reply;
	KQML_KMSG_LIST *kmsg_list, *temp;
	kqml_message *outmsg;
	char *reply_with, multiple_msg;

	/* ensure that we have a handler */
	if (hfunction == NULL)
	{
		error_reply (kmsg, "Can't handle performative");
		return ;
	}

	reply_with = _kqml_get_reply_with_field (kmsg);

				/* apply the handler function */
	_kqml_dbg_msg (KQML_DBG_VERBOSE) 
		("Handling message %s.", kqml_get_performative (kmsg));

	{
		int save_connection;

		save_connection = kmsg->connection;
		reply_type = KQML_REPLY_ERROR;
		reply = hfunction (kqml_get_field (":content", kmsg, NULL, NULL), 
			kmsg, &reply_type);
		kmsg_list = kqml_wrapper (reply_type, reply);
		kmsg->connection = save_connection;
	}

	/* prepare reply */
	if (! reply_with || kmsg_list == KQML_HANDLER_NONE) 
	{
		return ;
	}

	if (kmsg_list == KQML_HANDLER_ERROR)
	{
		error_reply (kmsg, "handler encountered error");
		return ;
	}

	multiple_msg = (kmsg_list->next != kmsg_list);
	temp = kmsg_list;
	do
	{
		outmsg = temp->kmsg;
		temp = temp->next;
		kqml_prep_outmsg (kmsg, outmsg);
		_kqml_send_xdr_msg (outmsg);
	}
	while (temp != kmsg_list);
	kqml_free_kmsg_list (kmsg_list);

	if (multiple_msg)
	{
		if (kqml_stricmp (kqml_get_performative (kmsg_list->prev->kmsg), 
				"eos") != 0)
		{
			/* If the last message is not an eos, we will send an eos */
			outmsg = kqml_build_msg ("eos", NULL, NULL, NULL, NULL);
			if (outmsg == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Could not prepare eos.");
				error_reply (kmsg, "not enough memory");
			}
			else
			{
				kqml_prep_outmsg (kmsg, outmsg);
				_kqml_send_xdr_msg (outmsg);
				kqml_free_msg (outmsg);
			}
		}
	}

	return;
}

void
_kqml_handler (kqml_message *kmsg)
{
	handler hfunction;
	char *perf; 

	/* Process the performative */
	/* IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT */
	/* All the handlers should save the connection field before
		invoking user functions and restore it after returning from
		the user function */

	perf = kqml_get_performative (kmsg);

	/* identify the performative */
	hfunction = _kqml_find_handler (perf);

	if (kqml_stricmp (perf, "next") == 0) /* processing NEXT */
	{
		process_next (kmsg);
	}
	else if (kqml_stricmp (perf, "rest") == 0) /* processing REST */
	{
		process_rest (kmsg);
	}
	else if (kqml_stricmp (perf, "discard") == 0) 
		/* processing DISCARD */
	{
		process_discard (kmsg);
	}
	else if (kqml_stricmp (perf, "monitor") == 0) /* processing MONITOR */
	{
		process_monitor (kmsg);
	}
	else
	{
		if (hfunction == NULL)
		{
			error_reply (kmsg, "not ready for performative.");
		}
		else if (kqml_stricmp (perf, "generator") == 0) 
		{
			process_generator (hfunction, kmsg);
		}
		else /* default processing */
		{
			process_and_reply (hfunction, kmsg);
		}
	}

	kqml_free_msg (kmsg);

	return;
}

static void
error_reply (kqml_message *kmsg, char *message)
{
	if (kmsg)
	{
		if (_kqml_get_reply_with_field (kmsg))
		{
			kqml_message *outmsg;
			
			outmsg = kqml_build_msg ("error", message, NULL, NULL, NULL);

			if (outmsg != NULL)
			{
				kqml_prep_outmsg (kmsg, outmsg);
				_kqml_send_xdr_msg (outmsg);
				kqml_free_msg (outmsg);
			}
		}
	}
}

void
kqml_prep_outmsg (kqml_message *kmsg, kqml_message *outmsg)
{
	char *sender, *receiver, *reply_with, *kmsg_id, *agent;
	int sender_len, receiver_len, reply_with_len, kmsg_id_len;

	sender = kqml_get_field (":sender", kmsg, &sender_len, NULL);
	receiver = kqml_get_field (":receiver", kmsg, &receiver_len, NULL);
	reply_with = kqml_get_field (":reply-with", kmsg, &reply_with_len, NULL);
	kmsg_id = kqml_get_field (KMSG_ID_KEYWORD, kmsg, &kmsg_id_len, NULL);

	/* Set the connection field */
	outmsg->connection = kmsg->connection;

	/* Set the :sender field of the outgoing message */
	if (receiver)
	{
		/* We may have multiple names. So, use the name that was
			there in the incoming message. If there was no name,
			use kqml_agent_name. */
		kqml_put_field (":sender", outmsg, receiver, receiver_len);
	}
	else
	{
		/* otherwise we will put our name */
		kqml_put_field (":sender", outmsg, kqml_agent_name, -1);
	}

	/* if the :receiver field of the outgoing message is not set */
	agent = kqml_get_field (":receiver", outmsg, NULL, NULL);
	if (agent == NULL)
	{
		/* Set the :receiver field of the outgoing message */
		if (sender)
		{
			kqml_put_field (":receiver", outmsg, sender, sender_len);
			agent = sender;
		}
	}

	if (_kqml_im_an_ans && agent)
	{
		/* 
		 * If we do not know the connection and if we are an ANS and
		 * if the :receiver field of the message is defined, we also
		 * set the address of the receiver or another ANS which
		 * might know the address of the receiver in the message.
		 */

		char *l_agent, *protocol, *addr_info;

		if (_kqml_ans_lookup (agent, &l_agent, &protocol, &addr_info))
		{
			char *address;
			int len;

			len = strlen (l_agent) + strlen (protocol) + strlen (addr_info) +
				5;
			address = malloc (len);
			if (address != NULL)
			{
				char *temp;

				temp = address;
				*temp++ = '(';
				temp = memccpy (temp, l_agent, 0, len);
				*(temp - 1) = ' ';
				temp = memccpy (temp, protocol, 0, len);
				*(temp - 1) = ' ';
				temp = memccpy (temp, addr_info, 0, len);
				*(temp - 1) = ')';
				*temp = '\0';

				_kqml_put_field_nd (":kqml-address", outmsg, address, 
					len, KQML_TYPE_SEXPR);
			}
		}
	}

	/* Set the :in-reply-to field of the outgoing message */
	if (reply_with)
	{
		kqml_put_field (":in-reply-to", outmsg, reply_with, reply_with_len);
	}

	/* Set the :KmSg_iD field of the outgoing message */
	if (kmsg_id)
	{
		kqml_put_field (KMSG_ID_KEYWORD, outmsg, kmsg_id, kmsg_id_len);
	}
}

KQML_KMSG_LIST *kqml_build_kmsg_list (KQML_KMSG_LIST **reply_msg, 
	kqml_message *kmsg)
{
	KQML_KMSG_LIST *kmsg_list;

	kmsg_list = malloc (sizeof (KQML_KMSG_LIST));
	if (kmsg_list == NULL)
	{
		return KQML_HANDLER_ERROR;
	}
	else
	{
		kmsg_list->kmsg = kmsg;
	}

	if (reply_msg == NULL)
	{
		kmsg_list->next =
			kmsg_list->prev = kmsg_list;
	}
	else if (*reply_msg == KQML_HANDLER_NONE || *reply_msg == KQML_HANDLER_ERROR)
	{
		*reply_msg =
			kmsg_list->next = 
			kmsg_list->prev = kmsg_list;
	}
	else
	{
		kmsg_list->next = *reply_msg;
		kmsg_list->prev = (*reply_msg)->prev;
		(*reply_msg)->prev->next = kmsg_list;
		(*reply_msg)->prev = kmsg_list;

		kmsg_list = *reply_msg;
	}

	return kmsg_list;
}

void kqml_free_kmsg_list (KQML_KMSG_LIST *reply)
{
	KQML_KMSG_LIST *temp, *temp2;

	if (reply == NULL)
		return;

	temp = reply;
	do
	{
		kqml_free_msg (temp->kmsg);
		temp2 = temp;
		temp = temp->next;
		free (temp2);
	}
	while (temp != reply);

	return;
}
