/*
 * send_pars_msg.c
 *
 * Write a parsed KQML message to an appropriate destination
 * used only by ROUTER
 * Richard Fritzson & Robin McEntire
 * Written: December 1992
 * Revised: October  1993
 *
 */

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <memory.h>

#if defined (__KQML_MACH_SUN__)
extern char *sys_errlist[];
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

/*
 * rwrite - a reliable write
 *
 * Like write, but doesn't suffer from EINTRs.
 * Also, if length < 0, rwrite will perform an strlen on the buffer.
 * 
 * Return: N > 0, number of bytes written
 * 	   N < 0, error (see errno)
 *
 */

static int rwrite (void *handle, char *buffer, u_int length);
static int rfwrite (void *handle, char *buffer, u_int length);

/* Note:
	rwrite and rfwrite should *NEVER* be passed a length > MAX_INT
	as this will interfere with error checking
*/
static int rwrite (void *handle, char *buffer, u_int length)
{
	int fd, write_size, size_to_write = length;

	fd = (int)handle;
	while (size_to_write > 0) 
	{
		write_size = write (fd, buffer, size_to_write);
		if (write_size == -1)
		{
			if (errno != EINTR)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("send_parsed_msg: %s.", sys_errlist[errno]);
				return -1;
			}
		}
		else 
		{
			buffer += write_size;
			size_to_write -= write_size;
		}
	}

	return length;
}

static int rfwrite (void *handle, char *buffer, u_int length)
{
	int write_size, size_to_write = length;
	FILE *fp;

	fp = (FILE *)handle;
	while (size_to_write > 0) 
	{
		write_size = fwrite (buffer, 1, size_to_write, fp);
		if (write_size == -1)
		{
			if (errno != EINTR)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("send_parsed_msg: %s.", sys_errlist[errno]);
				return -1;
			}
		}
		else 
		{
			buffer += write_size;
			size_to_write -= write_size;
		}
	}

	return length;
}

int
send_parsed_msg (kqml_message *kmsg, int con, FILE *stream)
{
	int index, field_count, value_len, temp;
	kqml_expression_type value_type;
	char *field, *value;
	char *kmsg_begin = "(";
	char *kmsg_end = ")";
	char tempStr[41]; /* Enough for very large ints atleast upto 2^128 */
	int (*write_func)(void *handle, char *string, u_int string_len);
	void *handle;

	if (stream)
	{
		write_func = rfwrite;
		handle = stream;
	}
	else
	{
		write_func = rwrite;
		handle = (void *)con;
	}

	if (write_func (handle, kmsg_begin, strlen (kmsg_begin)) <= -1 )
	{
		return (-1);
	}

	field = kqml_get_performative (kmsg);
	if (write_func (handle, field, strlen(field)) <= -1)
	{
		return (-1);
	}

	field_count = kqml_get_field_count (kmsg);
	for (index = 0; index < field_count; index++)
	{
		kqml_get_ith_field (index, kmsg, &field, &value, &value_len, &value_type);

		if (value_type == KQML_TYPE_UNKNOWN)
		{
			value_type = _kqml_check_type (value, value_len, &value, &value_len);
		}

		if (value_len > 0)
		{
			/* write a space between fields */
			if (write_func (handle, " ", 1) <= -1)
			{
				return (-1);
			}

			/* write the keyword */
			if (write_func (handle, field, strlen (field)) <= -1)
			{
				return (-1);
			}

			/* write a space between fields */
			if (write_func (handle, " ", 1) <= -1)
			{
				return (-1);
			}

			if (value_type == KQML_TYPE_STRING)
			{
				if (value[0] != '"')
				{
					if (write_func (handle, "\"", 1) <= -1)
					{
						return (-1);
					}
				}
			}
			else if (value_type == KQML_TYPE_BYTE)
			{
				if (value[0] != '#')
				{
					tempStr[0] = '#';
					temp = kqml_digit_to_str(value_len, tempStr+1);
					tempStr[temp+1] = '"';
					
					if (write_func (handle, tempStr, temp+2) <= -1)
					{
						return (-1);
					}
				}
			}

  		   /* write the key's value */
			if (write_func (handle, value, value_len) <= -1)
			{
				return (-1);
			}

			if (value_type == KQML_TYPE_STRING && value[0] != '"')
			{
				if (write_func (handle, "\"", 1) <= -1)
				{
					return (-1);
				}
			}
		}
	}

	if (write_func (handle, kmsg_end, strlen(kmsg_end)) <= -1)
	{
		return (-1);
	}

	if (stream)
	{
		fflush (stream);
	}

	return (0);
}
